//
// libsemigroups - C++ library for semigroups and monoids
// Copyright (C) 2026 James D. Mitchell
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

// This file exists just to include the headers below which comprise
// essentially all the functionality for generic matrices in libsemigroups.

#ifndef LIBSEMIGROUPS_MATRIX_HPP_
#define LIBSEMIGROUPS_MATRIX_HPP_

#include "is-matrix.hpp"
#include "matrix-adapters.hpp"
#include "matrix-class.hpp"
#include "matrix-exceptions.hpp"
#include "matrix-fmt.hpp"
#include "matrix-helpers.hpp"
#include "matrix-make.hpp"
#include "matrix-view.hpp"

#endif  // LIBSEMIGROUPS_MATRIX_HPP_
